/* *****************************************************************
 *
 * Copyright 2016 Samsung Electronics All Rights Reserved.
 *
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * *****************************************************************/
#include "utils.h"

#include "oic_malloc.h"
#include "experimental/logger.h"
#include "ocstack.h"
#include "ocpayload.h"
#include "pmutility.h"
#include "srmutility.h"
#include "cacommonutil.h"
#include "aclresource.h"
#include "ocpayloadcbor.h"
#include "acl_logging.h"
#include "experimental/payload_logging.h"
#include "utlist.h"

#define TAG "OIC_CLOUD_ACL_ID"

/**
 * ACL Id parse from received response
 *
 * @param[in] ctx       context
 * @param[out] data     data required to external application
 * @param[in] response  peer response
 * @return  OCStackResult application result
 */
static OCStackResult getAclIdFromResponse(void *ctx, void **data, OCClientResponse *response)
{
    OC_UNUSED(ctx);

    VERIFY_NON_NULL_RET(response, TAG, "NULL response", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(response->payload, TAG, "NULL payload response", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(data, TAG, "NULL data", OC_STACK_INVALID_PARAM);

    char *aclid = NULL;

    if (!OCRepPayloadGetPropString((const OCRepPayload *)response->payload, OC_RSRVD_ACL_ID, &aclid))
    {
        OIC_LOG_V(ERROR, TAG, "Can't get: %s", OC_RSRVD_ACL_ID);
        return OC_STACK_MALFORMED_RESPONSE;
    }

    OIC_LOG_V(INFO, TAG, "Received Acl id = %s", aclid);

    *data = aclid;
    return OC_STACK_OK;
}

/**
 * ACL Id by device request received data handler
 *
 * @param[in] ctx       context
 * @param[out] data     data required to external application
 * @param[in] response  peer response
 * @return  OCStackResult application result
 */
static OCStackResult handleGetAclIdByDeviceResponse(void *ctx, void **data,
        OCClientResponse *response)
{
    return getAclIdFromResponse(ctx, data, response);
}

/**
 * ACL Id create request received data handler
 *
 * @param[in] ctx       context
 * @param[out] data     data required to external application
 * @param[in] response  peer response
 * @return  OCStackResult application result
 */
static OCStackResult handleAclIdCreateResponse(void *ctx, void **data, OCClientResponse *response)
{
    return getAclIdFromResponse(ctx, data, response);
}

OCStackResult OCCloudGetAclIdByDevice(void *ctx,
                                      const char *deviceId,
                                      const char *cloudUri,
                                      OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH] = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(deviceId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s?%s=%s",
                                       cloudUri, OC_RSRVD_ACL_ID_URL, OC_RSRVD_DEVICE_ID, deviceId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }

    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, handleGetAclIdByDeviceResponse, NULL);

    return OCDoResource(NULL, OC_REST_GET, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

OCStackResult OCCloudAclIdCreate(void *ctx,
                                 const char *ownerId,
                                 const char *deviceId,
                                 const char *cloudUri,
                                 OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH] = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(ownerId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(deviceId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s?%s=%s&%s=%s",
                                       cloudUri, OC_RSRVD_ACL_ID_URL,
                                       OC_RSRVD_OWNER_ID, ownerId, OC_RSRVD_DEVICE_ID, deviceId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, handleAclIdCreateResponse, NULL);

    return OCDoResource(NULL, OC_REST_PUT, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

OCStackResult OCCloudAclIdDelete(void *ctx,
                                 const char *aclId,
                                 const char *cloudUri,
                                 OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH]  = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s?%s=%s", cloudUri, OC_RSRVD_ACL_ID_URL,
                                       OC_RSRVD_ACL_ID, aclId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, NULL, NULL);

    return OCDoResource(NULL, OC_REST_DELETE, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

/**
 * ACL Get Info received data handler
 *
 * @param[in] ctx       context
 * @param[out] data     data required to external application
 * @param[in] response  peer response
 * @return  OCStackResult application result
 */
static OCStackResult handleAclGetInfoResponse(void *ctx, void **data, OCClientResponse *response)
{
    OCStackResult result = OC_STACK_OK;
    uint8_t *cbor = NULL;
    size_t size   = 0;
    OicSecAcl_t *acl = NULL;

    OC_UNUSED(ctx);
    OC_UNUSED(data);

    VERIFY_NON_NULL_RET(response, TAG, "NULL response", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(response->payload, TAG, "NULL pyload response", OC_STACK_INVALID_PARAM);

    result = OCConvertPayload(response->payload, OC_FORMAT_CBOR, &cbor, &size);
    if (result != OC_STACK_OK)
    {
        OIC_LOG(ERROR, TAG, "Can't convert OCPayload to cbor");
        goto exit;
    }

    acl = CBORPayloadToCloudAcl(cbor, size);
    if (NULL == acl)
    {
        OIC_LOG(ERROR, TAG, "Can't parse CBOR payload");
        goto exit;
    }

    OIC_LOG_ACL(INFO, acl);

    if (NULL == acl->aces)
    {
        OIC_LOG(WARNING, TAG, "NULL aces received. Database update is not required");
        OIC_LOG(WARNING, TAG, "Assume that it is correct behavior");
        goto exit;
    }

    result = InstallACL(acl);
    if (result != OC_STACK_OK)
    {
        OIC_LOG(ERROR, TAG, "Can't update ACL resource");
    }

    *data = acl;
exit:
    OICFree(cbor);
    return result;
}

OCStackResult OCCloudAclIndividualGetInfo(void *ctx,
        const char *aclId,
        const char *cloudUri,
        OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH]  = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s/%s", cloudUri, OC_RSRVD_ACL_ID_URL,
                                       aclId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, handleAclGetInfoResponse, NULL);

    return OCDoResource(NULL, OC_REST_GET, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

static bool aceValidate(cloudAce_t *aces)
{
    cloudAce_t *ace = NULL;
    size_t acllist_count = 0;
    LL_FOREACH(aces, ace)
    {
        size_t reslist_count = 0;
        OicSecRsrc_t *res = NULL;
        LL_FOREACH( ace->resources, res)
        {
            if (NULL == res->href)
            {
                OIC_LOG_V(WARNING, TAG, "%s: wrong href of the resourse", __func__);
                continue;
            }
            reslist_count++;
        }
        if(0 == reslist_count)
        {
            OIC_LOG_V(WARNING, TAG, "%s: there is no resourse", __func__);
            return false;
        }

        acllist_count++;
    }
    if(0 == acllist_count)
    {
        OIC_LOG_V(WARNING, TAG, "%s: there is no acl", __func__);
        return false;
    }
    return true;
}

OCStackResult OCCloudAclIndividualAclUpdate(void *ctx,
        const char *aclId,
        const cloudAce_t *aces,
        const char *cloudUri,
        OCCloudResponseCB callback)
{
    size_t dimensions[MAX_REP_ARRAY_DEPTH] = { 0 };
    char uri[MAX_URI_LENGTH]  = { 0 };

    int i = 0, j = 0;

    OCRepPayload **helperPayload  = NULL;
    OCRepPayload **helperPayload2 = NULL;
    cloudAce_t *ace = NULL;
    size_t acllist_count = 0;
    OCCallbackData cbData;

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aces, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if(!aceValidate((cloudAce_t*)aces))
    {
        OIC_LOG_V(WARNING, TAG, "%s: invalid aces", __func__);
        return OC_STACK_INVALID_PARAM;
    }

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s/%s", cloudUri, OC_RSRVD_ACL_ID_URL,
                                       aclId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCRepPayload *payload = OCRepPayloadCreate();
    if (!payload)
    {
        OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for payload");
        goto no_memory;
    }

    acllist_count = 0;
    //code below duplicates LL_COUNT, implemented in newer version of utlist.h
    {
        cloudAce_t *ace = (cloudAce_t *)aces;
        while (ace)
        {
            ace = ace->next;
            acllist_count++;
        }
    }

    helperPayload = (OCRepPayload **)OICCalloc(acllist_count, sizeof(OCRepPayload *));
    if (!helperPayload)
    {
        OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload");
        goto no_memory;
    }

    i = 0;
    ace = NULL;
    LL_FOREACH((cloudAce_t *)aces, ace)
    {
        OCRepPayload *acePayload = OCRepPayloadCreate();
        if (!acePayload)
        {
            OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload[i]");
            goto no_memory;
        }
        helperPayload[i++] = acePayload;

        char *uuid = NULL;
        if (OC_STACK_OK != ConvertUuidToStr(&ace->subjectuuid, &uuid))
        {
            OIC_LOG(ERROR, TAG, "Can't convert subjectuuid to string");
        }

        OCRepPayloadSetPropString(acePayload, OC_RSRVD_SUBJECT_UUID, (const char *)uuid);
        OCRepPayloadSetPropInt(acePayload, OC_RSRVD_SUBJECT_TYPE, ace->stype);
        OCRepPayloadSetPropInt(acePayload, OC_RSRVD_PERMISSION_MASK, ace->permission);

        OICFree(uuid);

        size_t reslist_count = 0;
        //code below duplicates LL_COUNT, implemented in newer version of utlist.h
        {
            OicSecRsrc_t *res = ace->resources;
            while (res)
            {
                res = res->next;
                reslist_count++;
            }
        }

        helperPayload2 = (OCRepPayload **)OICCalloc(reslist_count, sizeof(OCRepPayload *));
        if (!helperPayload2)
        {
            OIC_LOG_V(ERROR, TAG, "%s: Can't allocate memory", __func__);
            goto no_memory;
        }

        j = 0;
        OicSecRsrc_t *res = NULL;

        LL_FOREACH(ace->resources, res)
        {
            OCRepPayload *aceResourcePayload = OCRepPayloadCreate();
            if (!aceResourcePayload)
            {
                OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload2[j]");
                goto no_memory;
            }
            helperPayload2[j++] = aceResourcePayload;

            OCRepPayloadSetPropString(aceResourcePayload, OC_RSRVD_HREF, res->href);

            dimensions[0] = res->typeLen;
            OCRepPayloadSetStringArray(aceResourcePayload, OC_RSRVD_RESOURCE_TYPE,
                                       (const char **)res->types, dimensions);

            dimensions[0] = res->interfaceLen;
            OCRepPayloadSetStringArray(aceResourcePayload, OC_RSRVD_INTERFACE,
                                       (const char **)res->interfaces, dimensions);
        }
        dimensions[0] = reslist_count;
        OCRepPayloadSetPropObjectArray(acePayload, OC_RSRVD_RESOURCES,
                                       (const OCRepPayload **)helperPayload2, dimensions);
    }
    dimensions[0] = acllist_count;
    OCRepPayloadSetPropObjectArray(payload, OC_RSRVD_ACCESS_CONTROL_LIST,
                                   (const OCRepPayload **)helperPayload, dimensions);

    fillCallbackData(&cbData, ctx, callback, NULL, NULL);

    OIC_LOG(DEBUG, TAG, "Next payload created:");
    OIC_LOG_PAYLOAD(DEBUG, (OCPayload *)payload);

    return OCDoResource(NULL, OC_REST_POST, uri, NULL, (OCPayload *)payload,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
no_memory:
    if (helperPayload2)
    {
        for (int k = 0; k < j; k++) OCRepPayloadDestroy(helperPayload2[k]);
        OICFree(helperPayload2);
    }
    if (helperPayload)
    {
        for (int k = 0; k < i; k++) OCRepPayloadDestroy(helperPayload[k]);
        OICFree(helperPayload);
    }
    OCRepPayloadDestroy(payload);
    return OC_STACK_NO_MEMORY;
}

OCStackResult OCCloudAclIndividualAceUpdate(void *ctx,
        const char *aclId,
        const char *aceId,
        const cloudAce_t *aces,
        const char *cloudUri,
        OCCloudResponseCB callback)
{
    size_t dimensions[MAX_REP_ARRAY_DEPTH] = { 0 };
    char uri[MAX_URI_LENGTH]  = { 0 };
    cloudAce_t *ace;

    int i = 0, j = 0;

    OCRepPayload **helperPayload  = NULL;
    OCRepPayload **helperPayload2 = NULL;
    size_t acllist_count;
    OCCallbackData cbData;

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aceId, TAG, "NULL aceId", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aces, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if(!aceValidate((cloudAce_t*)aces))
    {
        OIC_LOG_V(WARNING, TAG, "%s: invalid aces", __func__);
        return OC_STACK_INVALID_PARAM;
    }

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s/%s?%s=%s", cloudUri,
                                       OC_RSRVD_ACL_ID_URL, aclId,
                                       OC_RSRVD_ACE_ID, aceId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCRepPayload *payload = OCRepPayloadCreate();
    if (!payload)
    {
        OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for payload");
        goto no_memory;
    }

    acllist_count = 1;

    helperPayload = (OCRepPayload **)OICCalloc(acllist_count, sizeof(OCRepPayload *));
    if (!helperPayload)
    {
        OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload");
        goto no_memory;
    }

    i = 0;
    ace = NULL;

    LL_FOREACH((cloudAce_t *)aces, ace)
    {
        OCRepPayload *acePayload = OCRepPayloadCreate();
        if (!acePayload)
        {
            OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload[i]");
            goto no_memory;
        }
        helperPayload[i++] = acePayload;

        char *uuid = NULL;
        if (OC_STACK_OK != ConvertUuidToStr(&ace->subjectuuid, &uuid))
        {
            OIC_LOG(ERROR, TAG, "Can't convert subjectuuid to string");
        }

        OCRepPayloadSetPropString(acePayload, OC_RSRVD_SUBJECT_UUID, (const char *)uuid);
        OCRepPayloadSetPropInt(acePayload, OC_RSRVD_SUBJECT_TYPE, ace->stype);
        OCRepPayloadSetPropInt(acePayload, OC_RSRVD_PERMISSION_MASK, ace->permission);

        OICFree(uuid);

        size_t reslist_count = 0;
        //code below duplicates LL_COUNT, implemented in newer version of utlist.h
        {
            OicSecRsrc_t *res = ace->resources;
            while (res)
            {
                res = res->next;
                reslist_count++;
            }
        }

        helperPayload2 = (OCRepPayload **)OICCalloc(reslist_count, sizeof(OCRepPayload *));
        if (!helperPayload2)
        {
            goto no_memory;
        }

        j = 0;
        OicSecRsrc_t *res = NULL;

        LL_FOREACH(ace->resources, res)
        {
            OCRepPayload *aceResourcePayload = OCRepPayloadCreate();
            if (!aceResourcePayload)
            {
                OIC_LOG_V(DEBUG, TAG, "Can't allocate memory for helperPayload2[j]");
                goto no_memory;
            }
            helperPayload2[j++] = aceResourcePayload;

            OCRepPayloadSetPropString(aceResourcePayload, OC_RSRVD_HREF, res->href);

            dimensions[0] = res->typeLen;
            OCRepPayloadSetStringArray(aceResourcePayload, OC_RSRVD_RESOURCE_TYPE,
                                       (const char **)res->types, dimensions);

            dimensions[0] = res->interfaceLen;
            OCRepPayloadSetStringArray(aceResourcePayload, OC_RSRVD_INTERFACE,
                                       (const char **)res->interfaces, dimensions);
        }
        dimensions[0] = reslist_count;
        OCRepPayloadSetPropObjectArray(acePayload, OC_RSRVD_RESOURCES,
                                       (const OCRepPayload **)helperPayload2, dimensions);
    }
    dimensions[0] = acllist_count;
    OCRepPayloadSetPropObjectArray(payload, OC_RSRVD_ACCESS_CONTROL_LIST,
                                   (const OCRepPayload **)helperPayload, dimensions);

    fillCallbackData(&cbData, ctx, callback, NULL, NULL);

    OIC_LOG(DEBUG, TAG, "Next payload created:");
    OIC_LOG_PAYLOAD(DEBUG, (OCPayload *)payload);

    return OCDoResource(NULL, OC_REST_POST, uri, NULL, (OCPayload *)payload,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
no_memory:
    if (helperPayload2)
    {
        for (int k = 0; k < j; k++) OCRepPayloadDestroy(helperPayload2[k]);
        OICFree(helperPayload2);
    }
    if (helperPayload)
    {
        for (int k = 0; k < i; k++) OCRepPayloadDestroy(helperPayload[k]);
        OICFree(helperPayload);
    }
    OCRepPayloadDestroy(payload);
    return OC_STACK_NO_MEMORY;
}



OCStackResult OCCloudAclAcesDelete(void *ctx,
                                   const char *aclId,
                                   const char *cloudUri,
                                   OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH]  = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s/%s", cloudUri, OC_RSRVD_ACL_ID_URL,
                                       aclId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, NULL, NULL);

    return OCDoResource(NULL, OC_REST_DELETE, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

OCStackResult OCCloudAclIndividualAceDelete(void *ctx,
        const char *aclId,
        const char *aceId,
        const char *cloudUri,
        OCCloudResponseCB callback)
{
    char uri[MAX_URI_LENGTH]  = { 0 };

    VERIFY_NON_NULL_RET(cloudUri, TAG, "NULL endpoint", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aclId, TAG, "NULL input param", OC_STACK_INVALID_PARAM);
    VERIFY_NON_NULL_RET(aceId, TAG, "NULL aceId", OC_STACK_INVALID_PARAM);

    if (MAX_URI_LENGTH - 1 <= snprintf(uri, MAX_URI_LENGTH, "%s%s/%s?%s=%s", cloudUri,
                                       OC_RSRVD_ACL_ID_URL, aclId,
                                       OC_RSRVD_ACE_ID, aceId))
    {
        OIC_LOG_V(WARNING, TAG, "%s: uri truncated: %s", __func__, uri);
    }


    OCCallbackData cbData;
    fillCallbackData(&cbData, ctx, callback, NULL, NULL);

    return OCDoResource(NULL, OC_REST_DELETE, uri, NULL, NULL,
                        CT_ADAPTER_TCP, OC_LOW_QOS, &cbData, NULL, 0);
}

